import math

from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.forms import get_double_slider_form
from flask_monitoringdashboard.core.plot import scatter, get_figure, get_margin, get_layout, get_average_bubble_size
from flask_monitoringdashboard.core.info_box import get_plot_info
from flask_monitoringdashboard.core.utils import get_endpoint_details
from flask_monitoringdashboard.database import Request, session_scope
from flask_monitoringdashboard.database.count import count_users, count_versions_endpoint
from flask_monitoringdashboard.database.count_group import get_value
from flask_monitoringdashboard.database.data_grouped import get_two_columns_grouped
from flask_monitoringdashboard.database.endpoint import get_users
from flask_monitoringdashboard.database.versions import get_versions, get_first_requests
from flask_monitoringdashboard.views.details.time_version import format_version

TITLE = 'User-Focused Multi-Version Performance'

AXES_INFO = '''In this graph, the X-axis presents the versions that are used. The Y-axis presents
(a subset of) all unique users, as specified by "dashboard.config.group_by". You can use the slider
to select a subset of the all unique users.'''

CONTENT_INFO = '''A circle represents the median execution time of a unique user in a certain version.
A larger execution time is presented by a larger circle. With this graph you can find out whether
there is a difference in performance across users.'''


FORM_SUBTITLE = ['Users', 'Versions']
FORM_TITLE = 'Select the number of users and versions'


@blueprint.route('/endpoint/<endpoint_id>/version_user', methods=['GET', 'POST'])
@secure
def version_user(endpoint_id):
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, endpoint_id)

        slider_max = [count_users(db_session, endpoint_id), count_versions_endpoint(db_session, endpoint_id)]
        form = get_double_slider_form(slider_max, subtitle=FORM_SUBTITLE, title=FORM_TITLE)
        graph = version_user_graph(db_session, endpoint_id, form)
    return render_template('fmd_dashboard/graph-details.html', details=details, graph=graph, form=form,
                           title='{} for {}'.format(TITLE, details['endpoint']),
                           information=get_plot_info(AXES_INFO, CONTENT_INFO))


def version_user_graph(db_session, endpoint_id, form):
    """
    :param db_session: session for the database
    :param endpoint_id: the endpoint_id to filter the data on
    :param form: form for reducing the size of the graph
    :return: an HTML bubble plot
    """
    users = get_users(db_session, endpoint_id, form.get_slider_value(0))
    versions = get_versions(db_session, endpoint_id, form.get_slider_value(1))

    first_request = get_first_requests(db_session, endpoint_id)
    values = get_two_columns_grouped(db_session, Request.group_by, Request.endpoint_id == endpoint_id)
    data = [[get_value(values, (user, v)) for v in versions] for user in users]

    average = get_average_bubble_size(data)
    trace = [scatter(
        x=[format_version(v, get_value(first_request, v)) for v in versions],
        hovertext=['Time: {:,.1f}ms'.format(data[i][j]) for j in range(len(versions))],
        y=[users[i]] * len(versions),
        name=users[i],
        mode='markers',
        marker={
            'color': [get_color(users[i])] * len(versions),
            'size': [math.sqrt(d) for d in data[i]],
            'sizeref': average,
            'sizemode': 'area'
        }
    ) for i in range(len(users))]

    layout = get_layout(
        height=350 + 40 * len(trace),
        xaxis={'title': 'Versions', 'type': 'category'},
        yaxis={'title': 'Users', 'type': 'category', 'autorange': 'reversed'},
        margin=get_margin(b=200)
    )
    return get_figure(layout, trace)



import datetime

import jwt
from flask import json, jsonify

from flask_monitoringdashboard import blueprint, config
from flask_monitoringdashboard.core.utils import get_details
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.endpoint import get_endpoints
from flask_monitoringdashboard.database.request import get_data_between


@blueprint.route('/get_json_data', defaults={'time_from': 0})
@blueprint.route('/get_json_data/<time_from>')
@blueprint.route('/get_json_data/<time_from>/<time_to>')
def get_json_data_from(time_from, time_to=None):
    """
    The returned data is the data that is encrypted using a security token. This security token is set in the
    configuration.
    :param time_from: (optional) if specified, only the data-values after this date are returned.
                      input must be an timestamp value (utc) (= integer)
                      :param time_to: (optional) if specified, only the data-values before this date are returned.
                      input must be an timestamp value (utc) (= integer)
    :return: all entries from the database. (endpoint-table)
    """
    data = []
    try:
        with session_scope() as db_session:
            time1 = datetime.datetime.utcfromtimestamp(int(time_from))
            time2 = None
            if time_to:
                time2 = datetime.datetime.utcfromtimestamp(int(time_to))
            for entry in get_data_between(db_session, time1, time2):
                # nice conversion to json-object
                data.append({
                    'endpoint_id': entry.endpoint_id,
                    'duration': entry.duration,
                    'time_requested': str(entry.time_requested),
                    'version_requested': entry.version_requested,
                    'group_by': entry.group_by,
                    'ip': entry.ip
                })
        return jwt.encode({'data': json.dumps(data)}, config.security_token, algorithm='HS256')
    except ValueError as e:
        return 'ValueError: {}'.format(e)


@blueprint.route('/get_json_monitor_rules')
def get_json_monitor_rules():
    """
    The returned data is the data that is encrypted using a security token. This security token is set in the
    configuration.
    :return: all entries from the database (rules-table)
    """
    data = []
    try:
        with session_scope() as db_session:
            for entry in get_endpoints(db_session):
                # nice conversion to json-object
                data.append({
                    'name': entry.name,
                    'last_requested': str(entry.last_requested),
                    'monitor_level': entry.monitor_level,
                    'time_added': str(entry.time_added),
                    'version_added': entry.version_added
                })
            return jwt.encode({'data': json.dumps(data)}, config.security_token, algorithm='HS256')
    except ValueError as e:
        return 'ValueError: {}'.format(e)


@blueprint.route('/get_json_details')
def get_json_details():
    """
    Some details about the deployment, such as the current version, etc...
    :return: a json-object with the details.
    """
    with session_scope() as db_session:
        return jsonify(get_details(db_session))



import unittest

from flask_monitoringdashboard.core.profiler.util.stringHash import StringHash


class TestStringHash(unittest.TestCase):

    def test_stringhash(self):
        string_hash = StringHash()

        self.assertEqual(string_hash.hash('abc'), 0)
        self.assertEqual(string_hash.hash('def'), 1)
        self.assertEqual(string_hash.hash('abc'), 0)

    def test_unhash(self):
        string_hash = StringHash()
        self.assertEqual(string_hash.unhash(string_hash.hash('abc')), 'abc')
        self.assertRaises(ValueError, string_hash.unhash, 'unknown')


import threading
import time
import unittest

from flask_monitoringdashboard.core.profiler import start_thread_last_requested, start_performance_thread, \
    start_profiler_thread, start_profiler_and_outlier_thread
from flask_monitoringdashboard.database import Endpoint
from flask_monitoringdashboard.test.utils import NAME, set_test_environment, clear_db, add_fake_data, get_test_app


class TestProfiler(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    @staticmethod
    def wait_until_threads_finished(num_threads):
        while threading.active_count() > num_threads:
            time.sleep(0.001)

    def test_start_thread_last_requested(self):
        num_threads = threading.active_count()
        start_thread_last_requested(Endpoint(id=1, name=NAME))
        self.assertEqual(threading.active_count(), num_threads + 1)
        self.wait_until_threads_finished(num_threads)

    def test_start_performance_thread(self):
        with self.app.test_request_context():
            from flask import request
            request.environ['REMOTE_ADDR'] = '127.0.0.1'
            num_threads = threading.active_count()
            start_performance_thread(Endpoint(id=1, name=NAME), 1234)
            self.assertEqual(threading.active_count(), num_threads + 1)
            self.wait_until_threads_finished(num_threads)

    def test_start_profiler_thread(self):
        with self.app.test_request_context():
            from flask import request
            request.environ['REMOTE_ADDR'] = '127.0.0.1'
            num_threads = threading.active_count()
            thread = start_profiler_thread(Endpoint(id=1, name=NAME))
            self.assertEqual(threading.active_count(), num_threads + 1)
            thread.stop(1)
            self.wait_until_threads_finished(num_threads)

    def test_start_profiler_and_outlier_thread(self):
        with self.app.test_request_context():
            from flask import request
            request.environ['REMOTE_ADDR'] = '127.0.0.1'
            num_threads = threading.active_count()
            thread = start_profiler_and_outlier_thread(Endpoint(id=1, name=NAME))
            self.assertEqual(threading.active_count(), num_threads + 2)
            thread.stop(1)
            self.wait_until_threads_finished(num_threads)



"""
    This file can be executed for developing purposes.
    It is not used when the flask_monitoring_dashboard is attached to an existing flask application.
"""
import random

from flask import Flask


def create_app():
    import flask_monitoringdashboard as dashboard
    import time

    app = Flask(__name__)

    dashboard.config.version = '3.1'
    dashboard.bind(app)

    def f():
        time.sleep(2)
        time.sleep(1)

    @app.route('/endpoint')
    def endpoint():
        f()
        return 'Ok'

    return app


if __name__ == '__main__':
    create_app().run(debug=True, threaded=True)



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.forms import get_slider_form
from flask_monitoringdashboard.core.plot import get_layout, get_figure, boxplot
from flask_monitoringdashboard.core.info_box import get_plot_info
from flask_monitoringdashboard.core.utils import get_endpoint_details, simplify
from flask_monitoringdashboard.database import Request, session_scope
from flask_monitoringdashboard.database.count import count_users
from flask_monitoringdashboard.database.count_group import get_value
from flask_monitoringdashboard.database.data_grouped import get_user_data_grouped
from flask_monitoringdashboard.database.endpoint import get_users

TITLE = 'Per-User Performance'

AXES_INFO = '''The X-axis presents the execution time in ms. The Y-axis presents (a subset of) 
all unique users.'''

CONTENT_INFO = '''This graph shows a horizontal boxplot for the unique users that uses the application.
With this graph you can found out whether the performance is different across different users.'''


@blueprint.route('/endpoint/<endpoint_id>/users', methods=['GET', 'POST'])
@secure
def users(endpoint_id):
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, endpoint_id)
        form = get_slider_form(count_users(db_session, endpoint_id), title='Select the number of users')
    graph = users_graph(endpoint_id, form)

    return render_template('fmd_dashboard/graph-details.html', details=details, graph=graph, form=form,
                           title='{} for {}'.format(TITLE, details['endpoint']),
                           information=get_plot_info(AXES_INFO, CONTENT_INFO))


def users_graph(id, form):
    """
    Return an HTML box plot with a specific number of
    :param id: get the data for this endpoint only
    :param form: instance of SliderForm
    :return:
    """
    with session_scope() as db_session:
        users = get_users(db_session, id, form.get_slider_value())
        times = get_user_data_grouped(db_session, lambda x: simplify(x, 10), Request.endpoint_id == id)
        data = [boxplot(name=u, values=get_value(times, u)) for u in users]

    layout = get_layout(
        height=350 + 40 * len(data),
        xaxis={'title': 'Execution time (ms)'},
        yaxis={'type': 'category', 'title': 'User'}
    )
    return get_figure(layout, data)



"""
    This file contains all unit tests for the endpoint-table in the database. (Corresponding to the file:
    'flask_monitoringdashboard/database/function_calls.py')
    See info_box.py for how to run the test-cases.
"""

import time
import unittest

from flask_monitoringdashboard.database.request import get_avg_duration

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.count import count_requests
from flask_monitoringdashboard.database.endpoint import get_endpoint_by_name
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, REQUESTS, \
    TIMES, NAME, GROUP_BY, IP, ENDPOINT_ID


class TestRequest(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_add_request(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.request import add_request
        name2 = 'main2'
        execution_time = 1234
        self.assertNotEqual(NAME, name2, 'Both cannot be equal, otherwise the test will fail')
        with session_scope() as db_session:
            endpoint = get_endpoint_by_name(db_session, name2)
            self.assertEqual(count_requests(db_session, endpoint.id), 0)
            add_request(db_session, execution_time, endpoint.id, ip=IP)
            self.assertEqual(count_requests(db_session, endpoint.id), 1)

    def test_get_data_from(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.request import get_data_between
        size = 2
        first = len(TIMES) - size - 1
        with session_scope() as db_session:
            result = get_data_between(db_session, TIMES[-size - 2], TIMES[-1])
            for i in range(size):
                self.assertEqual(result[i].endpoint.name, NAME)
                self.assertEqual(result[i].duration, REQUESTS[first + i])
                self.assertEqual(result[i].time_requested, TIMES[first + i])
                self.assertEqual(result[i].group_by, GROUP_BY)
                self.assertEqual(result[i].ip, IP)

    def test_get_data(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.request import get_data
        from flask_monitoringdashboard import config
        with session_scope() as db_session:
            result = get_data(db_session)
            self.assertEqual(len(result), len(REQUESTS))
            for i in range(len(REQUESTS)):
                self.assertEqual(result[i].endpoint.name, NAME)
                self.assertEqual(result[i].duration, REQUESTS[i])
                self.assertEqual(result[i].time_requested, TIMES[i])
                self.assertEqual(result[i].group_by, GROUP_BY)
                self.assertEqual(result[i].version_requested, config.version)
                self.assertEqual(result[i].ip, IP)

    def test_get_versions(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard import config
        from flask_monitoringdashboard.database.versions import get_versions
        with session_scope() as db_session:
            result = get_versions(db_session)
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0], config.version)

    def test_get_endpoints(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_endpoints
        with session_scope() as db_session:
            result = get_endpoints(db_session)
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0].name, NAME)

    def test_get_date_of_first_request(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.request import get_date_of_first_request
        with session_scope() as db_session:
            self.assertEqual(get_date_of_first_request(db_session), int(time.mktime(TIMES[0].timetuple())))

    def test_get_avg_duration(self):
        with session_scope() as db_session:
            self.assertEqual(get_avg_duration(db_session, ENDPOINT_ID), 12000)


import datetime
import unittest

import jwt
from flask import json

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, get_test_app, \
    REQUESTS, NAME, GROUP_BY, IP, TIMES, test_admin_secure, test_post_data, ENDPOINT_ID


class TestExportData(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_download_csv(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'download-csv')

    def test_export_data(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'view-csv')

    def test_submit_test_results(self):
        """
            Submit some collect_performance data.
        """
        test_results = {'test_runs': [], 'endpoint_exec_times': []}
        test_results['test_runs'].append(
            {'name': 'test_1', 'exec_time': 50, 'time': str(datetime.datetime.utcnow()), 'successful': True, 'iter': 1})
        test_results['endpoint_exec_times'].append({'endpoint': 'main', 'exec_time': 30, 'test_name': 'test_1'})
        test_results['app_version'] = '1.0'
        test_results['travis_job'] = '133.7'
        test_post_data(self, 'submit-test-results', test_results)

    def test_get_json_data_from(self):
        """
            Test whether the response is as it should be.
        """
        from flask_monitoringdashboard import config
        with self.app.test_client() as c:
            result = c.get('dashboard/get_json_data').data
        decoded = jwt.decode(result, config.security_token, algorithms=['HS256'])
        data = json.loads(decoded['data'])
        self.assertEqual(len(data), len(REQUESTS))
        for row in data:
            self.assertEqual(row['endpoint_id'], ENDPOINT_ID)
            self.assertIn(row['duration'], REQUESTS)
            self.assertEqual(row['version_requested'], config.version)
            self.assertEqual(row['group_by'], GROUP_BY)
            self.assertEqual(row['ip'], IP)

    def test_get_json_endpoints(self):
        """
            Test whether the response is as it should be.
        """
        from flask_monitoringdashboard import config
        with self.app.test_client() as c:
            result = c.get('dashboard/get_json_monitor_rules').data
        decoded = jwt.decode(result, config.security_token, algorithms=['HS256'])
        data = json.loads(decoded['data'])
        self.assertEqual(len(data), 2)
        row = data[0]
        self.assertEqual(row['name'], NAME)
        self.assertEqual(row['last_requested'], str(TIMES[0]))
        self.assertEqual(row['monitor_level'], 1)
        self.assertEqual(row['version_added'], config.version)

    def test_get_json_details(self):
        """
            Test whether the response is as it should be.
        """
        with self.app.test_client() as c:
            result = c.get('dashboard/get_json_details').data
        data = json.loads(result)
        import pkg_resources
        self.assertEqual(data['dashboard-version'], pkg_resources.require("Flask-MonitoringDashboard")[0].version)



import datetime
import unittest

from flask_monitoringdashboard.database import session_scope, TestEndpoint
from flask_monitoringdashboard.database.count_group import get_latest_test_version, count_requests_group, \
    count_times_tested, count_requests_per_day
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, ENDPOINT_ID, \
    add_fake_test_runs, NAME, TIMES


class TestCountGroup(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    @staticmethod
    def add_test_data(db_session):
        add_fake_test_runs()
        db_session.add(TestEndpoint(endpoint_id=ENDPOINT_ID, test_id=1, duration=1234, app_version="1.0",
                                    travis_job_id="1", time_added=datetime.datetime.utcnow()))
        db_session.add(TestEndpoint(endpoint_id=ENDPOINT_ID, test_id=1, duration=2345, app_version="1.0",
                                    travis_job_id="1", time_added=datetime.datetime.utcnow()))

    def test_get_latest_test_version(self):
        with session_scope() as db_session:
            self.assertIsNone(get_latest_test_version(db_session))
            self.add_test_data(db_session)
            self.assertEqual(get_latest_test_version(db_session), "1.0")

    def test_count_requests_group(self):
        with session_scope() as db_session:
            self.assertEqual(count_requests_group(db_session), [(1, 5)])

    def test_count_times_tested(self):
        with session_scope() as db_session:
            self.assertEqual(count_times_tested(db_session), [])
            self.add_test_data(db_session)
            self.assertEqual(count_times_tested(db_session), [(NAME, 2)])

    def test_count_requests_per_day(self):
        with session_scope() as db_session:
            self.assertEqual(count_requests_per_day(db_session, []), [])

            if TIMES[0].date() == TIMES[len(TIMES)-1].date():  # must all be on the same day
                self.assertEqual(count_requests_per_day(db_session, [TIMES[0].date()]), [[(1, 5)]])



from flask import request
from flask_wtf import FlaskForm
from wtforms import SubmitField
from wtforms.fields.html5 import IntegerRangeField

from flask_monitoringdashboard.core.forms.slider import DEFAULT_SLIDER_VALUE


class DoubleSliderForm(FlaskForm):
    """
        Class for generating a slider that can be used to reduce the graph.
    """
    slider0 = IntegerRangeField()
    slider1 = IntegerRangeField()
    submit = SubmitField('Update')
    title = 'Select two numbers below for reducing the size of the graph'
    subtitle = ['Subtitle0', 'Subtitle1']

    def get_slider_value(self, index):
        """
        :return: the value from the slider
        """
        name = 'slider{}'.format(index)
        if name in request.form:
            return request.form[name]
        return self.start_value[index]

    def content(self):
        return '''
          <div class="row"><b class="col">{}</b></div>
          <div class="row">  
            <div class="col-sm-3"><span>{}</span></div>
            <div class="col-sm-3" style="text-align: center"><output id="outputValue0">{}</output></div>
            <div class="col-sm-3" style="text-align: right"><span>{}</span></div>
          </div>
          <div class="row">
            <div class="col-sm-9">{}</div>
          </div>
          <div class="row"><b class="col">{}</b></div>
          <div class="row">  
            <div class="col-sm-3"><span>{}</span></div>
            <div class="col-sm-3" style="text-align: center"><output id="outputValue1">{}</output></div>
            <div class="col-sm-3" style="text-align: right"><span>{}</span></div>
          </div>
          <div class="row">
            <div class="col-sm-9">{}</div>
            <div class="col-sm-3">{}</div>
          </div>'''.format(self.subtitle[0], self.min_value[0], self.start_value[0], self.max_value[0],
                           self.slider0(style="width: 100%;", oninput="outputValue0.value=this.value",
                                        min=self.min_value[0], max=self.max_value[0], value=self.start_value[0]),
                           self.subtitle[1], self.min_value[1], self.start_value[1], self.max_value[1],
                           self.slider1(style="width: 100%;", oninput="outputValue1.value=this.value",
                                        min=self.min_value[1], max=self.max_value[1], value=self.start_value[1]),
                           self.submit(class_="btn btn-primary btn-block"))


def get_double_slider_form(slider_max=(100, 100), title=None, subtitle=None):
    """
    Return a SliderForm with the range from 0 to slider_max
    :param slider_max: maximum value for the slider
    :param title: override the default title
    :param subtitle: override the default titles of the 2 sliders
    :return: a SliderForm with the range (0 ... slider_max)
    """
    form = DoubleSliderForm(request.form)
    form.min_value = (1, 1)
    form.max_value = slider_max
    if title:
        form.title = title
    if subtitle:
        form.subtitle = subtitle
    if 'slider0' in request.form:
        form.start_value = [request.form['slider0'], request.form['slider1']]
    else:
        form.start_value = [min(DEFAULT_SLIDER_VALUE, form.max_value[i]) for i in range(2)]
    return form



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.forms import get_slider_form
from flask_monitoringdashboard.core.plot import boxplot, get_figure, get_layout, get_margin
from flask_monitoringdashboard.core.info_box import get_plot_info
from flask_monitoringdashboard.core.utils import get_endpoint_details, simplify
from flask_monitoringdashboard.database import Request, session_scope
from flask_monitoringdashboard.database.count import count_versions_endpoint
from flask_monitoringdashboard.database.count_group import get_value
from flask_monitoringdashboard.database.data_grouped import get_version_data_grouped
from flask_monitoringdashboard.database.endpoint import to_local_datetime
from flask_monitoringdashboard.database.versions import get_first_requests

TITLE = 'Per-Version Performance'

AXES_INFO = '''The X-axis presents the execution time in ms. The Y-axis presents the versions 
that are used.'''

CONTENT_INFO = '''This graph shows a horizontal boxplot for the versions that are used. With this
graph you can found out whether the performance changes across different versions.'''


@blueprint.route('/endpoint/<endpoint_id>/versions', methods=['GET', 'POST'])
@secure
def versions(endpoint_id):
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, endpoint_id)
        form = get_slider_form(count_versions_endpoint(db_session, endpoint_id), title='Select the number of versions')

        graph = versions_graph(db_session, endpoint_id, form)
        return render_template('fmd_dashboard/graph-details.html', details=details, graph=graph,
                               title='{} for {}'.format(TITLE, details['endpoint']), form=form,
                               information=get_plot_info(AXES_INFO, CONTENT_INFO))


def format_version(version, first_used):
    """
    :param version: name of the version
    :param first_used: datetime object when the version was first used.
    :return: string that represents the version
    """
    if not first_used:
        return version
    return '{}<br>{}'.format(version, to_local_datetime(first_used).strftime('%Y-%m-%d %H:%M'))


def versions_graph(db_session, endpoint_id, form):
    times = get_version_data_grouped(db_session, lambda x: simplify(x, 10), Request.endpoint_id == endpoint_id)
    first_requests = get_first_requests(db_session, endpoint_id, form.get_slider_value())
    data = [boxplot(
                name=format_version(request.version_requested, get_value(first_requests, request.version_requested)),
                values=get_value(times, request.version_requested),
                marker={'color': get_color(request.version_requested)})
            for request in first_requests]

    layout = get_layout(
        height=350 + 40 * len(first_requests),
        xaxis={'title': 'Execution time (ms)'},
        yaxis={'type': 'category', 'title': 'Version', 'autorange': 'reversed'},
        margin=get_margin()
    )
    return get_figure(layout=layout, data=data)



from flask import request
from flask_wtf import FlaskForm
from wtforms import validators, SubmitField, PasswordField, StringField, SelectField

from .daterange import get_daterange_form
from .double_slider import get_double_slider_form
from .slider import get_slider_form

MONITOR_CHOICES = [
    (0, '0 - Disabled'),
    (1, '1 - Performance'),
    (2, '2 - Profiler'),
    (3, '3 - Profiler + Outliers')]


class Login(FlaskForm):
    """ Used for serving a login form. """
    name = StringField('Username', [validators.data_required()])
    password = PasswordField('Password', [validators.data_required()])
    submit = SubmitField('Login')


class RunTests(FlaskForm):
    """ Used for serving a login form on /{{ link }}/testmonitor. """
    submit = SubmitField('Run selected tests')


class MonitorLevel(FlaskForm):
    """ Used in the Rules page (per endpoint)"""
    monitor = SelectField('Monitor', coerce=int, choices=MONITOR_CHOICES)


def get_monitor_form(endpoint, default_value):
    """ Return a form with the endpoint as a variable """
    form = MonitorLevel(request.form)
    form.monitor.default = default_value
    form.monitor.label = endpoint
    form.process()
    return form



"""
    Main class for adding all route-functions to user_app.
    Setup requires only to import this file. All other imports are done in this file
"""
from flask import redirect, url_for
from flask.helpers import send_from_directory

from flask_monitoringdashboard import blueprint, loc
# Import more route-functions
from . import auth
from . import dashboard
from . import details
from . import export
from . import rules
from . import configuration
from . import testmonitor


@blueprint.route('/static/<path:filename>')
def static(filename):
    """
    Serve static files
    :param filename: filename in the /static file
    :return: content of the file
    """
    return send_from_directory(loc() + 'static', filename)


@blueprint.route('/')
def index():
    """
    Redirect to the default page
    """
    return redirect(url_for('dashboard.overview'))



import ast

from flask import render_template
from flask_paginate import get_page_args, Pagination

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.plot import boxplot, get_figure, get_layout, get_margin
from flask_monitoringdashboard.core.timezone import to_local_datetime
from flask_monitoringdashboard.core.utils import get_endpoint_details, simplify
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.count import count_outliers
from flask_monitoringdashboard.database.outlier import get_outliers_sorted, get_outliers_cpus

DATA_POINTS = 50


@blueprint.route('/endpoint/<endpoint_id>/outliers')
@secure
def outliers(endpoint_id):
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, endpoint_id)
        page, per_page, offset = get_page_args(page_parameter='page', per_page_parameter='per_page')
        table = get_outliers_sorted(db_session, endpoint_id, offset, per_page)
        for outlier in table:
            outlier.request.time_requested = to_local_datetime(outlier.request.time_requested)
        all_cpus = get_outliers_cpus(db_session, endpoint_id)
        graph = cpu_load_graph(all_cpus)

        total = count_outliers(db_session, endpoint_id)
        pagination = Pagination(page=page, per_page=per_page, total=total, format_number=True,
                                css_framework='bootstrap4', format_total=True, record_name='outliers')
    return render_template('fmd_dashboard/outliers.html', details=details, table=table, pagination=pagination,
                           title='Outliers for {}'.format(details['endpoint']), graph=graph)


def cpu_load_graph(all_cpus):
    count = 0  # some outliers have no CPU info
    values = []  # list of lists that stores the CPU info
    for cpu in all_cpus:
        if not cpu:
            continue
        x = ast.literal_eval(cpu)
        values.append(x)
        count += 1

    simplified = [simplify(x, DATA_POINTS) for x in zip(*values)]
    cores = []
    for i in range(len(simplified)):
        cores.append('CPU core %d:' % i)

    data = [boxplot(name=cores[idx], values=simplified[idx], marker={'color': get_color(core)})
            for idx, core in enumerate(cores)]

    layout = get_layout(
        height=150 + 40 * len(cores),
        xaxis={'title': 'CPU loads (%)'},
        yaxis={'type': 'category', 'autorange': 'reversed'},
        margin=get_margin(l=100, t=0)
    )
    return get_figure(layout=layout, data=data)



"""
    This is the main package for all test-cases.
    If you want to run all tests, then you can use this command:

        python -m unittest discover

    Another option is to run the main function at the bottom of this file.
"""
import unittest


def get_test_suite():
    """
        Finds all files that are located in a file named 'test_*.py', where * can be anything
    """
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('flask_monitoringdashboard.test', pattern='test_*.py')
    return test_suite



import unittest

from flask_monitoringdashboard.database import Endpoint
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, get_test_app, NAME


class TestMeasurement(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_init_measurement(self):
        from flask_monitoringdashboard.core.measurement import init_measurement
        with self.app.app_context():
            init_measurement()

    def test_add_decorator(self):
        from flask_monitoringdashboard.core.measurement import add_decorator
        with self.app.app_context():
            for i in range(4):
                self.assertEqual(add_decorator(Endpoint(name=NAME, monitor_level=i)), None)
            self.assertRaises(ValueError, add_decorator, Endpoint(name=NAME, monitor_level=-1))



import unittest

from flask_monitoringdashboard.test.utils import get_test_app


class TestFactory(unittest.TestCase):

    def test_factory(self):
        """
            Create multiple applications and verify that the app doesn't break.
        """
        get_test_app()
        get_test_app()



import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, get_test_app, NAME, \
    test_admin_secure, ENDPOINT_ID


class TestResult(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_result_heatmap(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'endpoint/{}/hourly_load'.format(ENDPOINT_ID))

    def test_result_time_per_version_per_user(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'endpoint/{}/version_user'.format(ENDPOINT_ID))

    def test_result_time_per_version_per_ip(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'endpoint/{}/version_ip'.format(ENDPOINT_ID))

    def test_result_time_per_version(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'endpoint/{}/versions'.format(ENDPOINT_ID))

    def test_result_time_per_user(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'endpoint/{}/users'.format(ENDPOINT_ID))

    def test_result_outliers(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'endpoint/{}/outliers'.format(ENDPOINT_ID))

    def test_profiler(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'endpoint/{}/profiler'.format(ENDPOINT_ID))

    def test_grouped_profiler(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'endpoint/{}/grouped-profiler'.format(ENDPOINT_ID))



import datetime

from flask import request

import flask_monitoringdashboard.views.export.csv
import flask_monitoringdashboard.views.export.json
from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.tested_endpoints import add_endpoint_hit
from flask_monitoringdashboard.database.tests import add_test_result, add_or_update_test


@blueprint.route('/submit-test-results', methods=['POST'])
def submit_test_results():
    """
    Endpoint for letting Travis submit its unit test performance results to the Dashboard.
    :return: nothing, 204 (No Content)
    """
    results = request.get_json()
    travis_job_id = -1
    if results['travis_job']:
        travis_job_id = results['travis_job']
    app_version = '-1'
    if 'app_version' in results:
        app_version = results['app_version']
    test_runs = results['test_runs']
    endpoint_hits = results['endpoint_exec_times']

    with session_scope() as db_session:
        for test_run in test_runs:
            time = datetime.datetime.strptime(test_run['time'], '%Y-%m-%d %H:%M:%S.%f')
            add_or_update_test(db_session, test_run['name'], test_run['successful'], time, app_version)
            add_test_result(db_session, test_run['name'], test_run['exec_time'], time, app_version,
                            travis_job_id, test_run['iter'])

        for endpoint_hit in endpoint_hits:
            add_endpoint_hit(db_session, endpoint_hit['endpoint'], endpoint_hit['exec_time'], endpoint_hit['test_name'],
                             app_version, travis_job_id)

    return '', 204



import inspect
import sys
import threading
import time
import traceback
from collections import defaultdict

from flask_monitoringdashboard import user_app, config
from flask_monitoringdashboard.core.logger import log
from flask_monitoringdashboard.core.profiler.util import order_histogram
from flask_monitoringdashboard.core.profiler.util.pathHash import PathHash
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.endpoint import update_last_accessed
from flask_monitoringdashboard.database.request import add_request
from flask_monitoringdashboard.database.stack_line import add_stack_line


FILENAME = 'flask_monitoringdashboard/core/measurement.py'
FILENAME_LEN = len(FILENAME)

class StacktraceProfiler(threading.Thread):
    """
    Used for profiling the performance per line code.
    This is used when monitoring-level == 2 and monitoring-level == 3
    """

    def __init__(self, thread_to_monitor, endpoint, ip, outlier_profiler=None):
        threading.Thread.__init__(self)
        self._keeprunning = True
        self._thread_to_monitor = thread_to_monitor
        self._endpoint = endpoint
        self._ip = ip
        self._duration = 0
        self._histogram = defaultdict(float)
        self._path_hash = PathHash()
        self._lines_body = []
        self._total = 0
        self._outlier_profiler = outlier_profiler

    def run(self):
        """
        Continuously takes a snapshot from the stacktrace (only the main-thread). Filters everything before the
        endpoint has been called (i.e. the Flask library).
        Directly computes the histogram, since this is more efficient for performance
        :return:
        """
        current_time = time.time()
        while self._keeprunning:
            newcurrent_time = time.time()
            duration = newcurrent_time - current_time
            current_time = newcurrent_time

            try:
                frame = sys._current_frames()[self._thread_to_monitor]
            except KeyError:
                log('Can\'t get the stacktrace of the main thread. Stopping StacktraceProfiler')
                break
            in_endpoint_code = False
            self._path_hash.set_path('')
            # filename, line number, function name, source code line
            for fn, ln, fun, line in traceback.extract_stack(frame):
                if self._endpoint.name == fun:
                    in_endpoint_code = True
                if in_endpoint_code:
                    key = (self._path_hash.get_path(fn, ln), fun, line)
                    self._histogram[key] += duration
                if len(fn) > FILENAME_LEN and fn[-FILENAME_LEN:] == FILENAME and fun == "wrapper":
                    in_endpoint_code = True
            if in_endpoint_code:
                self._total += duration

            elapsed = time.time() - current_time
            if config.sampling_period > elapsed:
                time.sleep(config.sampling_period - elapsed)

        self._on_thread_stopped()

    def stop(self, duration):
        self._duration = duration * 1000
        if self._outlier_profiler:
            self._outlier_profiler.stop()
        self._keeprunning = False

    def _on_thread_stopped(self):
        with session_scope() as db_session:
            update_last_accessed(db_session, endpoint_name=self._endpoint.name)
            request_id = add_request(db_session, duration=self._duration, endpoint_id=self._endpoint.id, ip=self._ip)
            self._lines_body = order_histogram(self._histogram.items())
            self.insert_lines_db(db_session, request_id)

        if self._outlier_profiler:
            self._outlier_profiler.add_outlier(request_id)

    def insert_lines_db(self, db_session, request_id):
        position = 0
        for code_line in self.get_funcheader():
            add_stack_line(db_session, request_id, position=position, indent=0, duration=self._duration,
                           code_line=code_line)
            position += 1

        for key, val in self._lines_body:
            path, fun, line = key
            fn, ln = self._path_hash.get_last_fn_ln(path)
            indent = self._path_hash.get_indent(path)
            duration = val * self._duration / self._total
            add_stack_line(db_session, request_id, position=position, indent=indent, duration=duration,
                           code_line=(fn, ln, fun, line))
            position += 1

    def get_funcheader(self):
        lines_returned = []
        try:
            fun = user_app.view_functions[self._endpoint.name]
        except AttributeError:
            fun = None
        if hasattr(fun, 'original'):
            original = fun.original
            fn = inspect.getfile(original)
            lines, ln = inspect.getsourcelines(original)
            count = 0
            for line in lines:
                lines_returned.append((fn, ln + count, 'None', line.strip()))
                count += 1
                if line.strip()[:4] == 'def ':
                    return lines_returned
        raise ValueError('Cannot retrieve the function header')



import argparse
import csv
import datetime
import math
import os
import time
from unittest import TestLoader

import requests

CONVERT_TO_MS = 1000

# Determine if this script was called normally or if the call was part of a unit test on Travis.
# When unit testing, only run one dummy test from the testmonitor folder and submit to a dummy url.
test_folder = os.getcwd()[:os.getcwd().find('Flask-MonitoringDashboard')] + \
              'Flask-MonitoringDashboard/flask_monitoringdashboard/test/views/testmonitor'
times = '1'
url = 'https://httpbin.org/post'
if 'TRAVIS_BUILD_DIR' in os.environ:
    if 'flask-dashboard/Flask-MonitoringDashboard' not in os.getenv('TRAVIS_BUILD_DIR'):
        parser = argparse.ArgumentParser(description='Collecting performance results from the unit tests of a project.')
        parser.add_argument('--test_folder', dest='test_folder', default='./',
                            help='folder in which the unit tests can be found (default: ./)')
        parser.add_argument('--times', dest='times', default=5,
                            help='number of times to execute every unit test (default: 5)')
        parser.add_argument('--url', dest='url', default=None,
                            help='url of the Dashboard to submit the performance results to')
        args = parser.parse_args()
        test_folder = args.test_folder
        times = args.times
        url = args.url

# Show the settings with which this script will run.
print('Starting the collection of performance results with the following settings:')
print('  - folder containing unit tests: ', test_folder)
print('  - number of times to run tests: ', times)
print('  - url to submit the results to: ', url)
if not url:
    print('The performance results will not be submitted.')

# Initialize result dictionary and logs.
data = {'test_runs': [], 'endpoint_exec_times': []}
home = os.path.expanduser("~")
log = open(home + '/start_endpoint_hits.log', 'w')
log.write('"time","endpoint"\n')
log.close()
log = open(home + '/finish_endpoint_hits.log', 'w')
log.write('"time","endpoint"\n')
log.close()
log = open(home + '/test_runs.log', 'w')
log.write('"start_time","stop_time","test_name"\n')

# Find the tests and execute them the specified number of times.
# Add the performance results to the result dictionary.
suites = TestLoader().discover(test_folder, pattern="*test*.py")
for iteration in range(int(times)):
    for suite in suites:
        for case in suite:
            for test in case:
                test_result = None
                start_time_stamp = str(datetime.datetime.utcnow())
                time_before = time.time()
                test_result = test.run(test_result)
                time_after = time.time()
                end_time_stamp = str(datetime.datetime.utcnow())
                log.write('"{}","{}","{}"\n'.format(start_time_stamp, end_time_stamp, str(test)))
                execution_time = (time_after - time_before) * CONVERT_TO_MS
                data['test_runs'].append(
                    {'name': str(test), 'exec_time': execution_time, 'time': str(datetime.datetime.utcnow()),
                     'successful': (test_result.wasSuccessful() if test_result else False), 'iter': iteration + 1})
log.close()

# Read and parse the log containing the test runs into an array for processing.
test_runs = []
with open(home + '/test_runs.log') as log:
    reader = csv.DictReader(log)
    for row in reader:
        test_runs.append((datetime.datetime.strptime(row["start_time"], "%Y-%m-%d %H:%M:%S.%f"),
                          datetime.datetime.strptime(row["stop_time"], "%Y-%m-%d %H:%M:%S.%f"),
                          row['test_name']))

# Read and parse the log containing the start of the endpoint hits into an array for processing.
start_endpoint_hits = []
with open(home + '/start_endpoint_hits.log') as log:
    reader = csv.DictReader(log)
    for row in reader:
        start_endpoint_hits.append([datetime.datetime.strptime(row["time"], "%Y-%m-%d %H:%M:%S.%f"),
                                    row['endpoint']])

# Read and parse the log containing the finish of the endpoint hits into an array for processing.
finish_endpoint_hits = []
with open(home + '/finish_endpoint_hits.log') as log:
    reader = csv.DictReader(log)
    for row in reader:
        finish_endpoint_hits.append((datetime.datetime.strptime(row["time"], "%Y-%m-%d %H:%M:%S.%f"),
                                     row['endpoint']))

# Analyze the two arrays containing the start and finish times of the endpoints that were hit by the tests.
# Calculate the execution time each of these endpoint calls took.
number_of_hits = len(finish_endpoint_hits)
for hit in range(number_of_hits):
    time_stamp_a, endpoint_name_a = start_endpoint_hits[hit]
    time_stamp_b, endpoint_name_b = finish_endpoint_hits[hit]
    for test_run in test_runs:
        start_time, stop_time, test_name = test_run
        if start_time <= time_stamp_b <= stop_time:
            exec_time = math.ceil((time_stamp_b - time_stamp_a).total_seconds() * CONVERT_TO_MS)
            data['endpoint_exec_times'].append(
                {'endpoint': endpoint_name_b, 'exec_time': exec_time, 'test_name': test_name})
            break

# Retrieve the current version of the user app that is being tested.
with open(home + '/app_version.log', 'r') as log:
    data['app_version'] = log.read()

# Add the current Travis Build Job number.
data['travis_job'] = os.getenv('TRAVIS_JOB_NUMBER')

# Send test results and endpoint_name/test_name combinations to the Dashboard if specified.
if url:
    if ('TRAVIS_BUILD_DIR' in os.environ and 'flask-dashboard/Flask-MonitoringDashboard' not in os.getenv(
            'TRAVIS_BUILD_DIR')) or 'httpbin.org' not in url:
        if url[-1] == '/':
            url += 'submit-test-results'
        else:
            url += '/submit-test-results'
    try:
        requests.post(url, json=data)
        print('Sent unit test results to the Dashboard at', url)
    except Exception as e:
        print('Sending unit test results to the dashboard failed:\n{}'.format(e))



import unittest

from flask_monitoringdashboard.core.rules import get_rules
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, get_test_app, NAME


class TestRules(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_rules(self):
        with self.app.app_context():
            self.assertEqual(len(get_rules()), 2)
            self.assertEqual(len(get_rules(NAME)), 1)
            self.assertEqual(get_rules('unknown'), [])



"""
    The following methods can be used for retrieving a HTML box with information:
      - get_plot_info: for information about the usage of a plot
      - get_rules_info: for information about the monitoring level.
"""

from flask_monitoringdashboard.core.forms import MONITOR_CHOICES

GRAPH_INFO = '''You can hover the graph with your mouse to see the actual values. You can also use 
the buttons at the top of the graph to select a subset of graph, scale it accordingly or save the graph
as a PNG image.'''


def b(s):
    return '<b>{}</b>'.format(s)


def p(s):
    return '<p>{}</p>'.format(s)


def get_plot_info(axes='', content=''):
    """
    :param axes: If specified, information about the axis
    :param content: If specified, information about the content
    :return: a String with information in HTML
    """

    information = b('Graph') + p(GRAPH_INFO)

    if axes:
        information += b('Axes') + p(axes)

    if content:
        information += b('Content') + p(content)

    return information


def get_rules_info():
    """ :return: a string with information in HTML """
    info = b(MONITOR_CHOICES[0][1]) + \
           p('When the monitoring-level is set to 0, you don\'t monitor anything about the performance of this '
             'endpoint. The only data that is stored is when the ' + b('endpoint is last requested.'))

    info += b(MONITOR_CHOICES[1][1]) + \
            p('When the monitoring-level is set to 1, you collect data when the endpoint is last requested, plus '
              'data about the ' + b('performance and utilization') + ' of this endpoint (as a black-box).')

    info += b(MONITOR_CHOICES[2][1]) + \
            p('When the monitoring-level is set to 2, you get all the functionality from 1, plus data about the ' +
              b('performance per line of code') + ' from all requests.')

    info += b(MONITOR_CHOICES[3][1]) + \
            p('When the monitoring-level is set to 3, you get all the functionality from 2, including ' + b('more data'
              ' if a request is an outlier.'))
    return info



import os
import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, add_fake_test_runs, \
    get_test_app, login, NAME, test_admin_secure, test_get_ok, test_get_redirect


class TestSetup(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_index(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_get_redirect(self, '')

    def test_static(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_get_ok(self, 'static/css/custom.css')

    def test_configuration(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'configuration')

    def test_rules(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'rules')

    def test_test_result(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        add_fake_test_runs()
        test_admin_secure(self, 'testmonitor/{}'.format(NAME))

    def test_build_performance(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        add_fake_test_runs()
        test_admin_secure(self, 'testmonitor/test_build_performance')

    def test_build_performance_endpoints(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        add_fake_test_runs()
        test_admin_secure(self, 'testmonitor/endpoint_build_performance')

    def test_monitor_rule(self):
        """
            Test whether it is possible to monitor a rule
        """
        data = {'name': NAME, 'value': 0}
        with self.app.test_client() as c:
            login(c)
            self.assertEqual(200, c.post('dashboard/rules', data=data).status_code)

    def test_collect_performance(self):
        """
            Tests the collect_performance script.
        """
        os.environ['TRAVIS'] = 'True'
        import flask_monitoringdashboard.collect_performance
        del os.environ['TRAVIS']
        self.assertGreater(len(flask_monitoringdashboard.collect_performance.data['test_runs']), 0)
